
import org.neo4j.driver.AuthTokens;
import org.neo4j.driver.Config;
import org.neo4j.driver.Driver;
import org.neo4j.driver.GraphDatabase;
import org.neo4j.driver.Record;
import org.neo4j.driver.Result;
import org.neo4j.driver.Session;
import org.neo4j.driver.internal.logging.JULogging;


import net.spy.memcached.AddrUtil;
import net.spy.memcached.MemcachedClient;

import static org.neo4j.driver.Values.parameters;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

public class genealogy {

	private final int EXPIRE_TIME = 60;
	private static String memcachedServer;
	private static String memcachedPort;
	
	final static String ADD_PERSON = "add_person";
	final static String ADD_CHILD = "add_child";
	final static String RELATIONSHIP = "relationship";
	final static String DESCENDANTS = "num_living_descendants";
	final static Logger LOGGER = Logger.getLogger(genealogy.class.getName());
	
	Driver driver;
	MemcachedClient memcachedClient;

	public genealogy(String uri, String user, String password) throws IOException {
		driver = GraphDatabase.driver(uri, AuthTokens.basic(user, password), Config.builder().withLogging(new JULogging(java.util.logging.Level.WARNING)).build());
		memcachedClient = new MemcachedClient(AddrUtil.getAddresses(memcachedServer+":"+memcachedPort));
	}

	public void addPerson(String name, String birthYear, String deathYear) {
		try (Session session = driver.session()) {
			try {
				// CYPHER query to add a node of type Person with name and birthYear as
				// properties
				String createPersonNode = " CREATE (person:Person {name: $name, birthYear: $birthYear})";
				session.writeTransaction(
						tx -> tx.run(createPersonNode, parameters("name", name, "birthYear", birthYear)));

				if (deathYear != null) {
					// Add the deathYear property only if it is supplied by the user in the command
					// line argument
					String setDeathYear = " MATCH (person { name: $name })" + " SET person.deathYear = $deathYear";
					session.writeTransaction(
							tx -> tx.run(setDeathYear, parameters("name", name, "deathYear", deathYear)));
				}

				System.out.println("SUCCESS");
			} catch (Exception e) {
				System.out.println("FAILURE");
			}

		}
	}

	public void addChild(String childName, String parentName) {
		try (Session session = driver.session()) {
			try {
				// CYPHER query to find relevant child and parent nodes.
				// If they exist then create a CHILD type relationship between them.
				// MERGE clause helps to merge an existing CHILD relationship with the new one
				// thereby having a single CHILD relationship between any two nodes.
				String addChildRelationship = "MATCH (child:Person {name: $childName}) "
						+ "MATCH (parent:Person {name: $parentName}) " + "MERGE (child)-[:CHILD]->(parent)";

				session.writeTransaction(tx -> tx.run(addChildRelationship,
						parameters("childName", childName, "parentName", parentName)));

				System.out.println("SUCCESS");
			} catch (Exception e) {
				System.out.println("FAILURE");
			}
		}
	}

	private void getRelationship(String person1, String person2) {
		//Memcache does not accept whitespace in keys - replacing with underscore
		String key1 = person1.replace(' ', '_') + "-" + person2.replace(' ', '_');
		String key2 = person2.replace(' ', '_') + "-" + person1.replace(' ', '_');

		// Searching for relationship in memcached
		if (memcachedClient.get(key1) != null) {
			System.out.println(memcachedClient.get(key1));
			System.out.println("Cached value");
		} else if (memcachedClient.get(key2) != null) {
			System.out.println(memcachedClient.get(key2));
			System.out.println("Cached value");
		} else {
			try (Session session = driver.session()) {
				// CYPHER query finds all relationships of type CHILD in all directions with
				// multiple hops between two Person nodes and returns the relationship count
				String findIfRelated = "MATCH path=shortestPath((p1:Person)-[:CHILD*]-(p2:Person)) "
									 + "WHERE p1.name = $person1 AND p2.name = $person2 "
									 + "RETURN count(path) as rel_exists";

				Result result = session.run(findIfRelated, parameters("person1", person1, "person2", person2));

				while (result.hasNext()) {
					// Storing results as key-value with keys as '<Name1>-<Name2>' and vice versa
					// and values as 'RELATED' or 'NOT RELATED'
					Record record = result.next();
					if (record.get("rel_exists").asInt() > 0) {
						System.out.println("RELATED");
						memcachedClient.add(key1, EXPIRE_TIME, "RELATED");
						memcachedClient.add(key2, EXPIRE_TIME, "RELATED");
					} else {
						System.out.println("NOT RELATED");
						memcachedClient.add(key1, EXPIRE_TIME, "NOT RELATED");
						memcachedClient.add(key2, EXPIRE_TIME, "NOT RELATED");
					}
				}
			}
		}
	}

	public void getLivingDescendantCount(String name) {
		//Searching for existing results in memcached
		if (memcachedClient.get(name.replace(' ','_')) != null) {
			System.out.println(memcachedClient.get(name.replace(' ','_')));
			System.out.println("Cached value");
		} else {
			try (Session session = driver.session()) {

				String getLivingDescendantCount = "MATCH (:Person {name:$name})<-[:CHILD*]-(n:Person) "
						+ "WHERE NOT EXISTS(n.deathYear) " + "RETURN COUNT(n) AS count_desc";

				Result result = session.run(getLivingDescendantCount, parameters("name", name));

				while (result.hasNext()) {
					Record record = result.next();
					System.out.println(record.get("count_desc").asInt());
					// Adding the results to the memcached client for 60 seconds
					memcachedClient.set(name.replace(' ','_'), EXPIRE_TIME, record.get("count_desc").asInt());
				}
			}
		}
	}

	public void close() {
		// Closing all open connections to the database and the memcached server
		driver.close();
		memcachedClient.shutdown();
	}

	public static void main(String... args) {
		if (args.length == 0 || args.length == 1) {
			System.out.println("Proper usage: java genealogy [operation] [arguments]");
			System.exit(0);
		}
		
		try {
			System.setProperty("net.spy.log.LoggerImpl","net.spy.memcached.compat.log.SunLogger");
			Logger.getLogger("net.spy.memcached").setLevel(Level.WARNING);	
			
			InputStream input = new FileInputStream("./config.properties");
			Properties prop = new Properties();
			prop.load(input);
			
			memcachedServer = prop.getProperty("memcached.server");
			memcachedPort = prop.getProperty("memcached.port");
			
			genealogy chart = new genealogy(prop.getProperty("db.url"), prop.getProperty("db.user"), prop.getProperty("db.password"));
			
			switch (args[0]) {

			case ADD_PERSON:
				if (args.length == 2) {
					System.out.println("Arguments expected : name birthYear [deathYear]");
				} else if (args.length == 3)
					chart.addPerson(args[1], args[2], null);
				else
					chart.addPerson(args[1], args[2], args[3]);
				break;

			case ADD_CHILD:
				if (args.length == 2) {
					System.out.println("Arguments expected : child_name parent_name ");
				} else
					chart.addChild(args[1], args[2]);
				break;

			case RELATIONSHIP:
				if (args.length == 2) {
					System.out.println("Arguments expected : name1 name2 ");
				} else
					try{
						chart.getRelationship(args[1], args[2]);
					}catch (Exception e) {
						chart.close();
					}
				break;

			case DESCENDANTS:
				try{
					chart.getLivingDescendantCount(args[1]);
				}catch(Exception e) {
					chart.close();
				}
				break;

			default:
				System.out.println("Invalid argument");
			}

			chart.close();
		} catch (IOException e) {
			e.printStackTrace();
		} catch(SecurityException e) {
			e.printStackTrace();
		}
		
	}
}